/*
 * Copyright (C) 2011 Koji Suzuki
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.1 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

`include "rtavr_defs.v"
// global defines : IOR_HAVE_PORTA , IOR_HAVE_PORTB
//                 , IOR_HAVE_PORTC, IOR_HAVE_INT0, IOR_HAVE_SREG_SP

`define IOR_DO_SYNC
`define IOR_ARBIT_BY_ADDR

//`define IOR_HAVE_PORTA
//`define IOR_HAVE_PORTB
//`define IOR_HAVE_PORTC
//`define IOR_HAVE_SPI
//`define IOR_HAVE_USART
//`define IOR_HAVE_TWI
//`define IOR_HAVE_CPI0
//`define IOR_HAVE_CPI1

//`define IOR_HAVE_INT0
//`define IOR_HAVE_TIM0
//`define IOR_HAVE_TIM0_OC0A
//`define IOR_HAVE_TIM0_OC0B
//`define IOR_HAVE_SREG_SP

// Module Name: rtavr_ior
// 
// Target Device:   xc3s200a-4ft256
// Product Version: ISE 12.4
// Design Goal:     Balanced

module rtavr_ior # (
		parameter CPUID = 0	// not used
) (
		input RESET,
		input WR,RD, 
		input [7:0] DI,
		input [5:0] ADDR,
		output [7:0] DO
`ifdef SYS_DRIVEN_CLK4PH
		, input CLK_0
		, input CLK_90
		, input CLK_180
		, input CLK_270
`else // SYS_DRIVEN_CLK2X
		, input CLK2X
		, input CLK
`endif
`ifdef IOR_HAVE_SREG_SP
		, input  [7:0] FLAGS_IN
		, input  [15:0] SP_IN
		, output [7:0] FLAGS_OUT
		, output [15:0] SP_OUT
`endif
		// external I/O
		, output [23:0] DDR
		, output [23:0] PORT
		, input  [23:0] PIN
		// interrupt handling
`ifdef IOR_HAVE_INT0
		, input  RST_INT0
		, output INT0
  `ifdef IOR_HAVE_INT1
		, input  RST_INT1
		, output INT1
  `endif
`endif
`ifdef IOR_HAVE_RAMWIN
		, output RAMW_WE
		, output RAMW_OE
		, output [12:0] RAMW_ADDR
		, input [7:0] RAMW_DO
`endif
`ifdef IOR_HAVE_NVM
		, output ROM_WP
		, output RAM_WP
`endif

`ifdef IOR_HAVE_TIM0
		, input  RST_TIM0_OVF
		, output TIM0_OVF
		, input  RST_TIM0_COMPA	
		, output TIM0_COMPA	
		, input  RST_TIM0_COMPB
		, output TIM0_COMPB
`endif
`ifdef IOR_HAVE_SPI
		, input  RST_SPI_INT
		, output SPI_INT
`endif
`ifdef IOR_HAVE_USART
		, output USART_TX_INT
		, output USART_UDRE_INT
		, output USART_RX_INT
		, input RST_USART_TX_INT
`endif
`ifdef IOR_HAVE_TWI
		, input  RST_TWI_SLAVE
		, output TWI_SLAVE
`endif
`ifdef IOR_HAVE_CPI1
		// external I/O
		, output [7:0] CPI1_TXD
		, output CPI1_TX
		, input CPI1_TX_ACK
		, output CPI1_IXS
		, input CPI1_IXS_ACK

		, input [7:0] CPI1_RXD
		, input CPI1_RX
		, output CPI1_RX_ACK
		, input CPI1_IXR
		, output CPI1_IXR_ACK

		// interrupt
		, input RST_CPI1_INT
		, output CPI1_INT
`endif
`ifdef IOR_DUALPORT
   `ifdef SUPPORT_LOCKING
		, input CPU0_LK_REQ
		, input CPU0_LK_REQ2X
		, output CPU0_LK_NAK
		, input [5:0] CPU0_LK_ADDR

		, input CPU1_LK_REQ
		, input CPU1_LK_REQ2X
		, output CPU1_LK_NAK
		, input [5:0] CPU1_LK_ADDR
   `endif
		, input CPU1_WR,CPU1_RD
		, input [7:0] CPU1_DI
		, input [5:0] CPU1_ADDR
		, output [7:0] CPU1_DO
	`ifdef IOR_HAVE_SREG_SP
		, input  [7:0] CPU1_FLAGS_IN
		, input  [15:0] CPU1_SP_IN
		, output [7:0] CPU1_FLAGS_OUT
		, output [15:0] CPU1_SP_OUT
	`endif
	`ifdef IOR_HAVE_INT0
		, input  CPU1_RST_INT0
		, output CPU1_INT0
	  `ifdef IOR_HAVE_INT1
		, input  CPU1_RST_INT1
		, output CPU1_INT1
	  `endif
	`endif
	`ifdef IOR_HAVE_RAMWIN
		, output CPU1_RAMW_WE
		, output CPU1_RAMW_OE
		, output [12:0] CPU1_RAMW_ADDR
		, input [7:0] CPU1_RAMW_DO
	`endif
	`ifdef IOR_HAVE_NVM
		, output CPU1_ROM_WP
		, output CPU1_RAM_WP
	`endif

	`ifdef IOR_HAVE_TIM0
		, input  CPU1_RST_TIM0_OVF
		, output CPU1_TIM0_OVF
		, input  CPU1_RST_TIM0_COMPA	
		, output CPU1_TIM0_COMPA	
		, input  CPU1_RST_TIM0_COMPB
		, output CPU1_TIM0_COMPB
	`endif
	`ifdef IOR_HAVE_SPI
		, input  CPU1_RST_SPI_INT
		, output CPU1_SPI_INT
	`endif
	`ifdef IOR_HAVE_USART
		, output CPU1_USART_TX_INT
		, output CPU1_USART_UDRE_INT
		, output CPU1_USART_RX_INT
		, input CPU1_RST_USART_TX_INT
	`endif
	`ifdef IOR_HAVE_TWI
		, input  CPU1_RST_TWI_SLAVE
		, output CPU1_TWI_SLAVE
	`endif
	`ifdef IOR_HAVE_CPI1
		// interrupt
		, input CPU1_RST_CPI1_INT
		, output CPU1_CPI1_INT
	`endif
	`ifdef IOR_HAVE_CPI0
		// interrupt
		, input RST_CPI0_INT
		, output CPI0_INT
		// interrupt
		, input CPU1_RST_CPI0_INT
		, output CPU1_CPI0_INT
	`endif
`endif // IOR_DUALPORT
		);
  
`ifdef SYS_DRIVEN_CLK4PH
    wire __CLK_90 =  1'b0;
    wire __CLK_270 =  1'b0;
`else // SYS_DRIVEN_CLK2X
    wire __CLK_90 = CLK;
    wire __CLK_270 = ~CLK;
`endif

  `ifdef IOR_HAVE_RAMWIN
	wire PS0_RAMW_OE;
	wire [12:0] PS0_RAMW_ADDR;
	wire [7:0] PS0_RAMW_DO;
  `endif

   wire  [7:0] DO_async;
  `ifdef IOR_DO_SYNC
      reg  [7:0] r_do;
  `ifdef IOR_HAVE_RAMWIN
      assign DO[7:0] = PS0_RAMW_OE ? PS0_RAMW_DO : r_do[7:0];
  `else
      assign DO[7:0] = r_do[7:0];
  `endif

      always @(`posedge_CLK_90)
      begin
	   if (~__CLK_90) r_do <= DO_async;
      end
  `else
      assign DO[7:0] = DO_async[7:0];
  `endif

    wire [23:0] PORT_IN;

`ifdef IOR_HAVE_SPI
    wire SPI_MASTER;
    wire SPI_SLAVE;
    wire SS_DDR;
    wire SS_IN;
    wire SCK_IN;
    wire MOSI_IN;
    wire MISO_IN;
`endif
`ifdef IOR_HAVE_USART
    wire XCK_DDR;
    wire XCK_IN;
    wire XCK_OUT;
    wire TX_OUT;
    wire RX_IN;
    wire USART_XCKEN;
    wire USART_TXEN;
`endif
`ifdef IOR_HAVE_TWI
    wire TWI_EN;
    wire SDA_OUT;
    wire SCL_OUT;
    wire SDA_IN;
    wire SCL_IN;
`endif

`ifdef IOR_HAVE_GPIOR0
      reg [7:0] gpior0 = 0;

      always @(`posedge_CLK_270)
	begin
	  if (RESET)
	    begin
		gpior0 <= 0;
	    end
	  else if (~__CLK_270)
	    begin
		if (WR & (ADDR == `GPIOR0_ADDR)) gpior0 <= DI;
  `ifdef IOR_DUALPORT
		else if (CPU1_WR & (CPU1_ADDR == `GPIOR0_ADDR))
			gpior0 <= CPU1_DI;
  `endif
	    end
	end
`endif // IOR_HAVE_GPIOR0
`ifdef IOR_HAVE_GPIOR1
      reg [7:0] gpior1 = 0;

      always @(`posedge_CLK_270)
	begin
	  if (RESET)
	    begin
		gpior1 <= 0;
	    end
	  else if (~__CLK_270)
	    begin
		if (WR & (ADDR == `GPIOR1_ADDR)) gpior1 <= DI;
  `ifdef IOR_DUALPORT
		else if (CPU1_WR & (CPU1_ADDR == `GPIOR1_ADDR))
			gpior1 <= CPU1_DI;
  `endif
	    end
	end
`endif // IOR_HAVE_GPIOR1
`ifdef IOR_HAVE_GPIOR2
      reg [7:0] gpior2 = 0;

      always @(`posedge_CLK_270)
	begin
	  if (RESET)
	    begin
		gpior2 <= 0;
	    end
	  else if (~__CLK_270)
	    begin
		if (WR & (ADDR == `GPIOR2_ADDR)) gpior2 <= DI;
  `ifdef IOR_DUALPORT
		else if (CPU1_WR & (CPU1_ADDR == `GPIOR2_ADDR))
			gpior2 <= CPU1_DI;
  `endif
	    end
	end
`endif // IOR_HAVE_GPIOR2

	wire [5:0] PS0_ADDR;
	wire [7:0] PS0_DI;
	wire [7:0] PS0_DO;
	wire [7:0] PS0_FLAGS_OUT;
	wire [7:0] PS0_FLAGS_IN;
	wire [15:0] PS0_SP_OUT;
	wire [15:0] PS0_SP_IN;
	wire [7:0] PS0_CPI_TXD;
	wire [7:0] PS0_CPI_RXD;

  rtavr_ior_ps #(.CPUID(0)) IOR_PS0 (.RESET(RESET)
  `ifdef SYS_DRIVEN_CLK4PH
		, .CLK_0(CLK_0)
		, .CLK_180(CLK_180)
		, .CLK_270(CLK_270)
  `else // SYS_DRIVEN_CLK2X
		, .CLK2X(CLK2X)
		, .CLK(CLK)
  `endif
		, .ADDR(PS0_ADDR[5:0]), .DI(PS0_DI[7:0]), .DO(PS0_DO[7:0])
		, .RD(PS0_RD), .WR(PS0_WR)
		, .ACK(PS0_ACK)
  `ifdef IOR_HAVE_INT0
		, .INT0(PS0_INT0)
		, .PIN_INT0(PS0_PIN_INT0)
		, .RST_INT0(PS0_RST_INT0)
    `ifdef IOR_HAVE_INT1
		, .INT1(PS0_INT1)
		, .PIN_INT1(PS0_PIN_INT1)
		, .RST_INT1(PS0_RST_INT1)
    `endif
  `endif
		// SREG(FLAGS) & SP
		, .FLAGS_OUT(PS0_FLAGS_OUT[7:0]) , .FLAGS_IN(PS0_FLAGS_IN[7:0])
		, .SP_OUT(PS0_SP_OUT[15:0]) , .SP_IN(PS0_SP_IN[15:0])
  `ifdef IOR_HAVE_RAMWIN
		, .RAMW_WE(PS0_RAMW_WE)
		, .RAMW_OE(PS0_RAMW_OE)
		, .RAMW_ADDR(PS0_RAMW_ADDR)
		, .RAMW_DO(PS0_RAMW_DO)
  `endif
  `ifdef IOR_HAVE_NVM
		, .ROM_WP(PS0_ROM_WP)
		, .RAM_WP(PS0_RAM_WP)
  `endif
  `ifdef IOR_HAVE_TIM0
		// INTERRUPTS 
		, .TIM0_OVF(PS0_TIM0_OVF)
		, .RST_TIM0_OVF(PS0_RST_TIM0_OVF)
		, .TIM0_COMPA(PS0_TIM0_COMPA)
		, .RST_TIM0_COMPA(PS0_RST_TIM0_COMPA)
		, .TIM0_COMPB(PS0_TIM0_COMPB)
		, .RST_TIM0_COMPB(PS0_RST_TIM0_COMPB)
		//
    `ifdef IOR_HAVE_TIM0_OC0A
                , .OC0A_EN(OC0A_EN)
                , .OC0A_OUT(OC0A_OUT)
    `endif
    `ifdef IOR_HAVE_TIM0_OC0B
                , .OC0B_EN(OC0B_EN)
                , .OC0B_OUT(OC0B_OUT)
    `endif
  `endif
  `ifdef IOR_HAVE_CPI0
		// external I/O
		, .CPI0_TXD(PS0_CPI_TXD[7:0])
		, .CPI0_RXD(PS0_CPI_RXD[7:0])
		, .CPI0_TX(PS0_CPI_TX)
		, .CPI0_TX_ACK(PS0_CPI_TX_ACK)
		, .CPI0_RX(PS0_CPI_RX)
		, .CPI0_RX_ACK(PS0_CPI_RX_ACK)
		, .CPI0_IXR(PS0_CPI_IXR)
		, .CPI0_IXR_ACK(PS0_CPI_IXR_ACK)
		, .CPI0_IXS(PS0_CPI_IXS)
		, .CPI0_IXS_ACK(PS0_CPI_IXS_ACK)

		// interrupt
		, .CPI0_INT(PS0_CPI_INT)
		, .RST_CPI0_INT(PS0_RST_CPI_INT)
  `endif
		);

		assign PS0_ADDR[5:0]	    = ADDR[5:0];
		assign PS0_DI[7:0] 	    = DI[7:0];
		assign PS0_RD		    = RD;
		assign PS0_WR		    = WR;
  `ifdef IOR_HAVE_INT0
		assign	INT0		    = PS0_INT0;
//		assign	PS0_PIN_INT0;
		assign	PS0_RST_INT0	    = RST_INT0;
    `ifdef IOR_HAVE_INT1
		assign	INT1		    = PS0_INT1;
//		assign	PS0_PIN_INT1;
		assign	PS0_RST_INT1	    = RST_INT1;
    `endif
  `endif
	// SREG(FLAGS) & SP
		assign	FLAGS_OUT[7:0]     = PS0_FLAGS_OUT[7:0];
		assign	PS0_FLAGS_IN[7:0]  = FLAGS_IN[7:0];
		assign	SP_OUT[15:0]	   = PS0_SP_OUT[15:0];
		assign	PS0_SP_IN[15:0]    = SP_IN[15:0];
  `ifdef IOR_HAVE_RAMWIN
		assign	RAMW_WE		   = PS0_RAMW_WE;
		assign	RAMW_OE		   = PS0_RAMW_OE;
		assign	RAMW_ADDR	   = PS0_RAMW_ADDR;
		assign	PS0_RAMW_DO	   = RAMW_DO;
  `endif
  `ifdef IOR_HAVE_NVM
		assign	ROM_WP		   = PS0_ROM_WP;
		assign	RAM_WP		   = PS0_RAM_WP;
  `endif
  `ifdef IOR_HAVE_TIM0
	// INTERRUPTS 
		assign	TIM0_OVF	   = PS0_TIM0_OVF;
		assign	PS0_RST_TIM0_OVF   = RST_TIM0_OVF;
		assign	TIM0_COMPA	   = PS0_TIM0_COMPA;
		assign	PS0_RST_TIM0_COMPA = RST_TIM0_COMPA;
		assign	TIM0_COMPB	   = PS0_TIM0_COMPB;
		assign	PS0_RST_TIM0_COMPB = RST_TIM0_COMPB;
  `endif

`ifdef IOR_DUALPORT
	wire [5:0] PS1_ADDR;
	wire [7:0] PS1_DI;
	wire [7:0] PS1_DO;
	wire [7:0] PS1_FLAGS_OUT;
	wire [7:0] PS1_FLAGS_IN;
	wire [15:0] PS1_SP_OUT;
	wire [15:0] PS1_SP_IN;
	wire [7:0] PS1_CPI_TXD;
	wire [7:0] PS1_CPI_RXD;
  `ifdef IOR_HAVE_RAMWIN
	wire [12:0] PS1_RAMW_ADDR;
	wire [7:0] PS1_RAMW_DO;
  `endif


  rtavr_ior_ps #(.CPUID(1)) IOR_PS1 (.RESET(RESET)
  `ifdef SYS_DRIVEN_CLK4PH
		, .CLK_0(CLK_0)
		, .CLK_180(CLK_180)
		, .CLK_270(CLK_270)
  `else // SYS_DRIVEN_CLK2X
		, .CLK2X(CLK2X)
		, .CLK(CLK)
  `endif
		, .ADDR(PS1_ADDR[5:0]), .DI(PS1_DI[7:0]), .DO(PS1_DO[7:0])
		, .RD(PS1_RD), .WR(PS1_WR)
		, .ACK(PS1_ACK)
  `ifdef IOR_HAVE_INT0
		, .INT0(PS1_INT0)
		, .PIN_INT0(PS1_PIN_INT0)
		, .RST_INT0(PS1_RST_INT0)
    `ifdef IOR_HAVE_INT1
		, .INT1(PS1_INT1)
		, .PIN_INT1(PS1_PIN_INT1)
		, .RST_INT1(PS1_RST_INT1)
    `endif
  `endif
		// SREG(FLAGS) & SP
		, .FLAGS_OUT(PS1_FLAGS_OUT[7:0]) , .FLAGS_IN(PS1_FLAGS_IN[7:0])
		, .SP_OUT(PS1_SP_OUT[15:0]) , .SP_IN(PS1_SP_IN[15:0])
  `ifdef IOR_HAVE_RAMWIN
		, .RAMW_WE(PS1_RAMW_WE)
		, .RAMW_OE(PS1_RAMW_OE)
		, .RAMW_ADDR(PS1_RAMW_ADDR)
		, .RAMW_DO(PS1_RAMW_DO)
  `endif
  `ifdef IOR_HAVE_NVM
		, .ROM_WP(PS1_ROM_WP)
		, .RAM_WP(PS1_RAM_WP)
  `endif
  `ifdef IOR_HAVE_TIM0
		// INTERRUPTS 
		, .TIM0_OVF(PS1_TIM0_OVF)
		, .RST_TIM0_OVF(PS1_RST_TIM0_OVF)
		, .TIM0_COMPA(PS1_TIM0_COMPA)
		, .RST_TIM0_COMPA(PS1_RST_TIM0_COMPA)
		, .TIM0_COMPB(PS1_TIM0_COMPB)
		, .RST_TIM0_COMPB(PS1_RST_TIM0_COMPB)
		//
    `ifdef IOR_HAVE_TIM0_OC0A
                , .OC0A_EN(PS1_OC0A_EN)
                , .OC0A_OUT(PS1_OC0A_OUT)
    `endif
    `ifdef IOR_HAVE_TIM0_OC0B
                , .OC0B_EN(PS1_OC0B_EN)
                , .OC0B_OUT(PS1_OC0B_OUT)
    `endif
  `endif
  `ifdef IOR_HAVE_CPI0
		// external I/O
		, .CPI0_TXD(PS1_CPI_TXD[7:0])
		, .CPI0_RXD(PS1_CPI_RXD[7:0])
		, .CPI0_TX(PS1_CPI_TX)
		, .CPI0_TX_ACK(PS1_CPI_TX_ACK)
		, .CPI0_RX(PS1_CPI_RX)
		, .CPI0_RX_ACK(PS1_CPI_RX_ACK)
		, .CPI0_IXR(PS1_CPI_IXR)
		, .CPI0_IXR_ACK(PS1_CPI_IXR_ACK)
		, .CPI0_IXS(PS1_CPI_IXS)
		, .CPI0_IXS_ACK(PS1_CPI_IXS_ACK)

		// interrupt
		, .CPI0_INT(PS1_CPI_INT)
		, .RST_CPI0_INT(PS1_RST_CPI_INT)
  `endif
		);

		assign PS1_ADDR[5:0]	    = CPU1_ADDR[5:0];
		assign PS1_DI[7:0] 	    = CPU1_DI[7:0];
		assign PS1_RD		    = CPU1_RD;
		assign PS1_WR		    = CPU1_WR;
  `ifdef IOR_HAVE_INT0
		assign	CPU1_INT0	    = PS1_INT0;
		assign	PS1_RST_INT0	    = CPU1_RST_INT0;
    `ifdef IOR_HAVE_INT1
		assign	CPU1_INT1	    = PS1_INT1;
		assign	PS1_RST_INT1	    = CPU1_RST_INT1;
    `endif
  `endif
	// SREG(FLAGS) & SP
		assign	CPU1_FLAGS_OUT[7:0] = PS1_FLAGS_OUT[7:0];
		assign	PS1_FLAGS_IN[7:0]  = CPU1_FLAGS_IN[7:0];
		assign	CPU1_SP_OUT[15:0]  = PS1_SP_OUT[15:0];
		assign	PS1_SP_IN[15:0]    = CPU1_SP_IN[15:0];
  `ifdef IOR_HAVE_RAMWIN
		assign	CPU1_RAMW_WE	   = PS1_RAMW_WE;
		assign	CPU1_RAMW_OE	   = PS1_RAMW_OE;
		assign	CPU1_RAMW_ADDR	   = PS1_RAMW_ADDR;
		assign	PS1_RAMW_DO	   = CPU1_RAMW_DO;
  `endif
  `ifdef IOR_HAVE_NVM
		assign	CPU1_ROM_WP	   = PS1_ROM_WP;
		assign	CPU1_RAM_WP	   = PS1_RAM_WP;
  `endif
  `ifdef IOR_HAVE_TIM0
	// INTERRUPTS 
		assign	CPU1_TIM0_OVF	   = PS1_TIM0_OVF;
		assign	PS1_RST_TIM0_OVF   = CPU1_RST_TIM0_OVF;
		assign	CPU1_TIM0_COMPA	   = PS1_TIM0_COMPA;
		assign	PS1_RST_TIM0_COMPA = CPU1_RST_TIM0_COMPA;
		assign	CPU1_TIM0_COMPB	   = PS1_TIM0_COMPB;
		assign	PS1_RST_TIM0_COMPB = CPU1_RST_TIM0_COMPB;
  `endif
  `ifdef IOR_HAVE_CPI0	// only SMP
		// external I/O
		assign	PS0_CPI_RXD[7:0]   = PS1_CPI_TXD[7:0];
		assign	PS0_CPI_TX_ACK	   = PS1_CPI_RX_ACK;
		assign	PS0_CPI_RX	   = PS1_CPI_TX;
		assign	PS0_CPI_IXS_ACK    = PS1_CPI_IXR_ACK;
		assign	PS0_CPI_IXR	   = PS1_CPI_IXS;

		assign	PS1_CPI_RXD[7:0]   = PS0_CPI_TXD[7:0];
		assign	PS1_CPI_TX_ACK	   = PS0_CPI_RX_ACK;
		assign	PS1_CPI_RX	   = PS0_CPI_TX;
		assign	PS1_CPI_IXS_ACK    = PS0_CPI_IXR_ACK;
		assign	PS1_CPI_IXR	   = PS0_CPI_IXS;

	// interrupt
		assign	CPI0_INT	   = PS0_CPI_INT;
		assign	PS0_RST_CPI_INT    = RST_CPI0_INT;
	// interrupt
		assign	CPU1_CPI0_INT	   = PS1_CPI_INT;
		assign	PS1_RST_CPI_INT    = CPU1_RST_CPI0_INT;
  `endif
`endif // IOR_DUALPORT

  `ifdef IOR_HAVE_USART
	wire WE_ubrrl = WR & (ADDR == `USART_ADDR);
	wire WE_ubrrh = WR & (ADDR == (`USART_ADDR + 1));
	wire WE_ucsrc = WR & (ADDR == (`USART_ADDR + 2));
	wire WE_ucsrb = WR & (ADDR == (`USART_ADDR + 3));
	wire WE_ucsra = WR & (ADDR == (`USART_ADDR + 4));
	wire WE_udr   = WR & (ADDR == (`USART_ADDR + 5));
	wire OE_udr   = RD & (ADDR == (`USART_ADDR + 5)); 
	wire [7:0] DO_ubrrl;
	wire [7:0] DO_ubrrh;
	wire [7:0] DO_ucsrc;
	wire [7:0] DO_ucsrb;
	wire [7:0] DO_ucsra;
	wire [7:0] DO_udr;
  `endif
  `ifdef IOR_HAVE_SPI
	wire WE_spdr = WR & (ADDR == `SPI_ADDR);
	wire WE_spsr = WR & (ADDR == (`SPI_ADDR + 1));
	wire WE_spcr = WR & (ADDR == (`SPI_ADDR + 2));
	wire OE_spdr = RD & (ADDR == `SPI_ADDR);
	wire OE_spsr = RD & (ADDR == (`SPI_ADDR + 1)); 
	wire [7:0] DO_spdr;
	wire [7:0] DO_spsr;
	wire [7:0] DO_spcr;
  `endif
  `ifdef IOR_HAVE_CPI1
	wire WE_cpcsr1 = WR & (ADDR == `CPI1_ADDR);
	wire WE_cpdr1  = WR & (ADDR == (`CPI1_ADDR + 1));
//	wire OE_cpcsr1 = RD & (ADDR == `CPI1_ADDR);
	wire OE_cpdr1  = RD & (ADDR == (`CPI1_ADDR + 1)); 
	wire [7:0] DO_cpcsr1;
	wire [7:0] DO_cpdr1;
  `endif
  `ifdef IOR_HAVE_PORTA
	wire WE_ddr_a   = WR & (ADDR == `PORTA_ADDR);
	wire WE_pin_a   = WR & (ADDR == `PORTA_ADDR +1);
	wire WE_port_a  = WR & (ADDR == `PORTA_ADDR +2);
	wire [7:0] DO_ddr_a;
	wire [7:0] DO_pin_a;
	wire [7:0] DO_port_a;
  `endif
  `ifdef IOR_HAVE_PORTB
	wire WE_ddr_b   = WR & (ADDR == `PORTB_ADDR);
	wire WE_pin_b   = WR & (ADDR == `PORTB_ADDR +1);
	wire WE_port_b  = WR & (ADDR == `PORTB_ADDR +2);
	wire [7:0] DO_ddr_b;
	wire [7:0] DO_pin_b;
	wire [7:0] DO_port_b;
  `endif
  `ifdef IOR_HAVE_PORTC
	wire WE_ddr_c   = WR & (ADDR == `PORTC_ADDR);
	wire WE_pin_c   = WR & (ADDR == `PORTC_ADDR +1);
	wire WE_port_c  = WR & (ADDR == `PORTC_ADDR +2);
	wire [7:0] DO_ddr_c;
	wire [7:0] DO_pin_c;
	wire [7:0] DO_port_c;
  `endif
  
  assign DO_async = ~RD ? 8'h00
        : PS0_ACK ? PS0_DO
  `ifdef IOR_HAVE_USART
	: (ADDR == `USART_ADDR) ?       DO_ubrrl
	: (ADDR == (`USART_ADDR + 1)) ? DO_ubrrh
	: (ADDR == (`USART_ADDR + 2)) ? DO_ucsrc
	: (ADDR == (`USART_ADDR + 3)) ? DO_ucsrb
	: (ADDR == (`USART_ADDR + 4)) ? DO_ucsra
	: (ADDR == (`USART_ADDR + 5)) ? DO_udr
  `endif
  `ifdef IOR_HAVE_SPI
	: (ADDR == `SPI_ADDR) ?       DO_spdr
	: (ADDR == (`SPI_ADDR + 1)) ? DO_spsr
	: (ADDR == (`SPI_ADDR + 2)) ? DO_spcr
  `endif
  `ifdef IOR_HAVE_CPI1
	: (ADDR == `CPI1_ADDR) ?       DO_cpcsr1
	: (ADDR == (`CPI1_ADDR + 1)) ? DO_cpdr1
  `endif
  `ifdef IOR_HAVE_PORTA
	: (ADDR == `PORTA_ADDR) ?     DO_ddr_a
	: (ADDR == `PORTA_ADDR +1) ?  DO_pin_a
	: (ADDR == `PORTA_ADDR +2) ?  DO_port_a
  `endif
  `ifdef IOR_HAVE_PORTB
	: (ADDR == `PORTB_ADDR) ?     DO_ddr_b
	: (ADDR == `PORTB_ADDR +1) ?  DO_pin_b
	: (ADDR == `PORTB_ADDR +2) ?  DO_port_b
  `endif
  `ifdef IOR_HAVE_PORTC
	: (ADDR == `PORTC_ADDR) ?     DO_ddr_c
	: (ADDR == `PORTC_ADDR +1) ?  DO_pin_c
	: (ADDR == `PORTC_ADDR +2) ?  DO_port_c
  `endif
  `ifdef IOR_HAVE_GPIOR0
	: (ADDR == `GPIOR0_ADDR) ?     gpior0
  `endif
  `ifdef IOR_HAVE_GPIOR1
	: (ADDR == `GPIOR1_ADDR) ?     gpior1
  `endif
  `ifdef IOR_HAVE_GPIOR2
	: (ADDR == `GPIOR2_ADDR) ?     gpior2
  `endif
	: 8'h00;

`ifdef IOR_DUALPORT
  `ifdef IOR_HAVE_USART
	wire CPU1_WE_ubrrl = CPU1_WR & (CPU1_ADDR == `USART_ADDR);
	wire CPU1_WE_ubrrh = CPU1_WR & (CPU1_ADDR == (`USART_ADDR + 1));
	wire CPU1_WE_ucsrc = CPU1_WR & (CPU1_ADDR == (`USART_ADDR + 2));
	wire CPU1_WE_ucsrb = CPU1_WR & (CPU1_ADDR == (`USART_ADDR + 3));
	wire CPU1_WE_ucsra = CPU1_WR & (CPU1_ADDR == (`USART_ADDR + 4));
	wire CPU1_WE_udr   = CPU1_WR & (CPU1_ADDR == (`USART_ADDR + 5));
	wire CPU1_OE_udr   = CPU1_RD & (CPU1_ADDR == (`USART_ADDR + 5)); 
  `endif
  `ifdef IOR_HAVE_SPI
	wire CPU1_WE_spdr = CPU1_WR & (CPU1_ADDR == `SPI_ADDR);
	wire CPU1_WE_spsr = CPU1_WR & (CPU1_ADDR == (`SPI_ADDR + 1));
	wire CPU1_WE_spcr = CPU1_WR & (CPU1_ADDR == (`SPI_ADDR + 2));
	wire CPU1_OE_spdr = CPU1_RD & (CPU1_ADDR == `SPI_ADDR);
	wire CPU1_OE_spsr = CPU1_RD & (CPU1_ADDR == (`SPI_ADDR + 1)); 
  `endif
  `ifdef IOR_HAVE_CPI1
	wire CPU1_WE_cpcsr1 = CPU1_WR & (CPU1_ADDR == `CPI1_ADDR);
	wire CPU1_WE_cpdr1  = CPU1_WR & (CPU1_ADDR == (`CPI1_ADDR + 1));
//	wire CPU1_OE_cpcsr1 = CPU1_RD & (CPU1_ADDR == `CPI1_ADDR);
	wire CPU1_OE_cpdr1  = CPU1_RD & (CPU1_ADDR == (`CPI1_ADDR + 1)); 
  `endif
  `ifdef IOR_HAVE_PORTA
	wire CPU1_WE_ddr_a   = CPU1_WR & (CPU1_ADDR == `PORTA_ADDR);
	wire CPU1_WE_pin_a   = CPU1_WR & (CPU1_ADDR == `PORTA_ADDR +1);
	wire CPU1_WE_port_a  = CPU1_WR & (CPU1_ADDR == `PORTA_ADDR +2);
  `endif
  `ifdef IOR_HAVE_PORTB
	wire CPU1_WE_ddr_b   = CPU1_WR & (CPU1_ADDR == `PORTB_ADDR);
	wire CPU1_WE_pin_b   = CPU1_WR & (CPU1_ADDR == `PORTB_ADDR +1);
	wire CPU1_WE_port_b  = CPU1_WR & (CPU1_ADDR == `PORTB_ADDR +2);
  `endif
  `ifdef IOR_HAVE_PORTC
	wire CPU1_WE_ddr_c   = CPU1_WR & (CPU1_ADDR == `PORTC_ADDR);
	wire CPU1_WE_pin_c   = CPU1_WR & (CPU1_ADDR == `PORTC_ADDR +1);
	wire CPU1_WE_port_c  = CPU1_WR & (CPU1_ADDR == `PORTC_ADDR +2);
  `endif
  
   wire  [7:0] CPU1_DO_async;
  `ifdef IOR_DO_SYNC
      reg  [7:0] r_do1;
  `ifdef IOR_HAVE_RAMWIN
      assign CPU1_DO[7:0] = PS1_RAMW_OE ? PS1_RAMW_DO : r_do1[7:0];
  `else
      assign CPU1_DO[7:0] = r_do1[7:0];
  `endif

      always @(`posedge_CLK_90)
      begin
	   if (~__CLK_90) r_do1 <= CPU1_DO_async;
      end
  `else
      assign CPU1_DO[7:0] = CPU1_DO_async[7:0];
  `endif

  assign CPU1_DO_async = ~CPU1_RD ? 8'h00
        : PS1_ACK ? PS1_DO
  `ifdef IOR_HAVE_USART
	: (CPU1_ADDR == `USART_ADDR) ?       DO_ubrrl
	: (CPU1_ADDR == (`USART_ADDR + 1)) ? DO_ubrrh
	: (CPU1_ADDR == (`USART_ADDR + 2)) ? DO_ucsrc
	: (CPU1_ADDR == (`USART_ADDR + 3)) ? DO_ucsrb
	: (CPU1_ADDR == (`USART_ADDR + 4)) ? DO_ucsra
	: (CPU1_ADDR == (`USART_ADDR + 5)) ? DO_udr
  `endif
  `ifdef IOR_HAVE_SPI
	: (CPU1_ADDR == `SPI_ADDR) ?       DO_spdr
	: (CPU1_ADDR == (`SPI_ADDR + 1)) ? DO_spsr
	: (CPU1_ADDR == (`SPI_ADDR + 2)) ? DO_spcr
  `endif
  `ifdef IOR_HAVE_CPI1
	: (CPU1_ADDR == `CPI1_ADDR) ?       DO_cpcsr1
	: (CPU1_ADDR == (`CPI1_ADDR + 1)) ? DO_cpdr1
  `endif
  `ifdef IOR_HAVE_PORTA
	: (CPU1_ADDR == `PORTA_ADDR) ?     DO_ddr_a
	: (CPU1_ADDR == `PORTA_ADDR +1) ?  DO_pin_a
	: (CPU1_ADDR == `PORTA_ADDR +2) ?  DO_port_a
  `endif
  `ifdef IOR_HAVE_PORTB
	: (CPU1_ADDR == `PORTB_ADDR) ?     DO_ddr_b
	: (CPU1_ADDR == `PORTB_ADDR +1) ?  DO_pin_b
	: (CPU1_ADDR == `PORTB_ADDR +2) ?  DO_port_b
  `endif
  `ifdef IOR_HAVE_PORTC
	: (CPU1_ADDR == `PORTC_ADDR) ?     DO_ddr_c
	: (CPU1_ADDR == `PORTC_ADDR +1) ?  DO_pin_c
	: (CPU1_ADDR == `PORTC_ADDR +2) ?  DO_port_c
  `endif
  `ifdef IOR_HAVE_GPIOR0
	: (CPU1_ADDR == `GPIOR0_ADDR) ?     gpior0
  `endif
  `ifdef IOR_HAVE_GPIOR1
	: (CPU1_ADDR == `GPIOR1_ADDR) ?     gpior1
  `endif
  `ifdef IOR_HAVE_GPIOR2
	: (CPU1_ADDR == `GPIOR2_ADDR) ?     gpior2
  `endif
	: 8'h00;

  `ifdef IOR_HAVE_SPI
	wire DP_WE_spdr		= (WE_spdr    | CPU1_WE_spdr );
	wire DP_WE_spsr		= (WE_spsr    | CPU1_WE_spsr );
	wire DP_WE_spcr		= (WE_spcr    | CPU1_WE_spcr );
	wire DP_OE_spdr		= (OE_spdr    | CPU1_OE_spdr );
	wire DP_OE_spsr		= (OE_spsr    | CPU1_OE_spsr );
	wire [7:0] DI_spi	= (WE_spdr | WE_spsr | WE_spcr ) ? DI : CPU1_DI;
  `endif
  `ifdef IOR_HAVE_USART
	wire DP_WE_ubrrl	= (WE_ubrrl   | CPU1_WE_ubrrl );
	wire DP_WE_ubrrh	= (WE_ubrrh   | CPU1_WE_ubrrh );
	wire DP_WE_ucsrc	= (WE_ucsrc   | CPU1_WE_ucsrc );
	wire DP_WE_ucsrb	= (WE_ucsrb   | CPU1_WE_ucsrb );
	wire DP_WE_ucsra	= (WE_ucsra   | CPU1_WE_ucsra );
	wire DP_WE_udr		= (WE_udr     | CPU1_WE_udr );
	wire DP_OE_udr		= (OE_udr     | CPU1_OE_udr );
	wire [7:0] DI_usart	= (WE_ubrrl | WE_ubrrh | WE_ucsrc
				 | WE_ucsrb | WE_ucsra | WE_udr ) ? DI : CPU1_DI;
  `endif
  `ifdef IOR_HAVE_CPI1
	wire DP_WE_cpcsr1	= (WE_cpcsr1  | CPU1_WE_cpcsr1 );
	wire DP_WE_cpdr1	= (WE_cpdr1   | CPU1_WE_cpdr1 );
	wire DP_OE_cpdr1	= (OE_cpdr1   | CPU1_OE_cpdr1 );
	wire [7:0] DI_cpi1	= (WE_cpcsr1 | WE_cpdr1) ? DI : CPU1_DI;
  `endif
  `ifdef IOR_HAVE_PORTA
	wire DP_WE_ddr_a	= (WE_ddr_a   | CPU1_WE_ddr_a );
	wire DP_WE_pin_a	= (WE_pin_a   | CPU1_WE_pin_a );
	wire DP_WE_port_a	= (WE_port_a  | CPU1_WE_port_a );
    `ifdef IOR_ARBIT_BY_ADDR
	wire [7:0] DI_ddr_a  =  (WE_ddr_a)  ? DI : CPU1_DI;
	wire [7:0] DI_pin_a  =  (WE_pin_a)  ? DI : CPU1_DI;
	wire [7:0] DI_port_a =  (WE_port_a) ? DI : CPU1_DI;
    `else
	wire [7:0] DI_pa	= (WE_ddr_a | WE_pin_a | WE_port_a) ? DI : CPU1_DI;
	wire [7:0] DI_ddr_a  = DI_pa;
	wire [7:0] DI_pin_a  = DI_pa;
	wire [7:0] DI_port_a = DI_pa;
    `endif
  `endif
  `ifdef IOR_HAVE_PORTB
	wire DP_WE_ddr_b	= (WE_ddr_b   | CPU1_WE_ddr_b );
	wire DP_WE_pin_b	= (WE_pin_b   | CPU1_WE_pin_b );
	wire DP_WE_port_b	= (WE_port_b  | CPU1_WE_port_b );
    `ifdef IOR_ARBIT_BY_ADDR
	wire [7:0] DI_ddr_b  =  (WE_ddr_b)  ? DI : CPU1_DI;
	wire [7:0] DI_pin_b  =  (WE_pin_b)  ? DI : CPU1_DI;
	wire [7:0] DI_port_b =  (WE_port_b) ? DI : CPU1_DI;
    `else
	wire [7:0] DI_pb	= (WE_ddr_b | WE_pin_b | WE_port_b) ? DI : CPU1_DI;
	wire [7:0] DI_ddr_b  = DI_pb;
	wire [7:0] DI_pin_b  = DI_pb;
	wire [7:0] DI_port_b = DI_pb;
    `endif
  `endif
  `ifdef IOR_HAVE_PORTC
	wire DP_WE_ddr_c	= (WE_ddr_c   | CPU1_WE_ddr_c );
	wire DP_WE_pin_c	= (WE_pin_c   | CPU1_WE_pin_c );
	wire DP_WE_port_c	= (WE_port_c  | CPU1_WE_port_c );
    `ifdef IOR_ARBIT_BY_ADDR
	wire [7:0] DI_ddr_c  =  (WE_ddr_c)  ? DI : CPU1_DI;
	wire [7:0] DI_pin_c  =  (WE_pin_c)  ? DI : CPU1_DI;
	wire [7:0] DI_port_c =  (WE_port_c) ? DI : CPU1_DI;
    `else
	wire [7:0] DI_pc	= (WE_ddr_c | WE_pin_c | WE_port_c) ? DI : CPU1_DI;
	wire [7:0] DI_ddr_c  = DI_pc;
	wire [7:0] DI_pin_c  = DI_pc;
	wire [7:0] DI_port_c = DI_pc;
    `endif
  `endif

  `ifdef IOR_HAVE_SPI
	wire DP_SPI_INT;
	assign SPI_INT		= DP_SPI_INT;
	assign CPU1_SPI_INT	= DP_SPI_INT;
	wire DP_RST_SPI_INT	= ( RST_SPI_INT | CPU1_RST_SPI_INT);
  `endif
  `ifdef IOR_HAVE_USART
	wire DP_USART_TX_INT;
	wire DP_USART_UDRE_INT;
	wire DP_USART_RX_INT;
	assign USART_TX_INT		= DP_USART_TX_INT;
	assign CPU1_USART_TX_INT	= DP_USART_TX_INT;
	assign USART_UDRE_INT		= DP_USART_UDRE_INT;
	assign CPU1_USART_UDRE_INT	= DP_USART_UDRE_INT;
	assign USART_RX_INT		= DP_USART_RX_INT;
	assign CPU1_USART_RX_INT	= DP_USART_RX_INT;
	wire DP_RST_USART_TX_INT = ( RST_USART_TX_INT | CPU1_RST_USART_TX_INT);
  `endif
  `ifdef IOR_HAVE_CPI1
	wire DP_CPI1_INT;
	assign CPI1_INT		= DP_CPI1_INT;
	assign CPU1_CPI1_INT	= DP_CPI1_INT;
	wire DP_RST_CPI1_INT	= ( RST_CPI1_INT | CPU1_RST_CPI1_INT);
  `endif

`endif // IOR_DUALPORT
`ifdef IOR_HAVE_SPI
    rtavr_ior_spi #(.BASE_ADDR(`SPI_ADDR)) i_spi(
		.RESET(RESET)
	`ifdef SYS_DRIVEN_CLK4PH
		, .CLK_180(CLK_180)
	`else // SYS_DRIVEN_CLK2X
		, .CLK2X(CLK2X)
		, .CLK(CLK)
	`endif
	`ifdef IOR_DUALPORT
		, .DI(DI_spi)
		, .WE_spdr(DP_WE_spdr)
		, .WE_spsr(DP_WE_spsr)
		, .WE_spcr(DP_WE_spcr)
		, .OE_spdr(DP_OE_spdr)
		, .OE_spsr(DP_OE_spsr)
		, .SPI_INT(DP_SPI_INT)
		, .RST_SPI_INT(DP_RST_SPI_INT)
	`else
		, .DI(DI)
		, .WE_spdr(WE_spdr)
		, .WE_spsr(WE_spsr)
		, .WE_spcr(WE_spcr)
		, .OE_spdr(OE_spdr)
		, .OE_spsr(OE_spsr)
		, .SPI_INT(SPI_INT)
		, .RST_SPI_INT(RST_SPI_INT)
	`endif
		, .DO_spdr(DO_spdr)
		, .DO_spsr(DO_spsr)
		, .DO_spcr(DO_spcr)
		, .SS_DDR(SS_DDR), .SS_IN(SS_IN)
		, .SCK_OUT(SCK_OUT), .SCK_IN(SCK_IN)
		, .MOSI_OUT(MOSI_OUT), .MOSI_IN(MOSI_IN)
		, .MISO_OUT(MISO_OUT), .MISO_IN(MISO_IN)
		, .SPI_MASTER(SPI_MASTER), .SPI_SLAVE(SPI_SLAVE)
		);
`endif
`ifdef IOR_HAVE_USART
    rtavr_ior_usart #(.BASE_ADDR(`USART_ADDR)) i_usart(
		.RESET(RESET)
	`ifdef SYS_DRIVEN_CLK4PH
		, .CLK_180(CLK_180)
	`else // SYS_DRIVEN_CLK2X
		, .CLK2X(CLK2X)
		, .CLK(CLK)
	`endif
	`ifdef IOR_DUALPORT
		, .DI(DI_usart)
		, .WE_ubrrl(DP_WE_ubrrl)
		, .WE_ubrrh(DP_WE_ubrrh)
		, .WE_ucsrc(DP_WE_ucsrc)
		, .WE_ucsrb(DP_WE_ucsrb)
		, .WE_ucsra(DP_WE_ucsra)
		, .WE_udr(DP_WE_udr)
		, .OE_udr(DP_OE_udr)
		, .TX_INT(DP_USART_TX_INT)
		, .UDRE_INT(DP_USART_UDRE_INT)
		, .RX_INT(DP_USART_RX_INT)
		, .RST_TX_INT(DP_RST_USART_TX_INT)
	`else
		, .DI(DI)
		, .WE_ubrrl(WE_ubrrl)
		, .WE_ubrrh(WE_ubrrh)
		, .WE_ucsrc(WE_ucsrc)
		, .WE_ucsrb(WE_ucsrb)
		, .WE_ucsra(WE_ucsra)
		, .WE_udr(WE_udr)
		, .OE_udr(OE_udr)
		, .TX_INT(USART_TX_INT)
		, .UDRE_INT(USART_UDRE_INT)
		, .RX_INT(USART_RX_INT)
		, .RST_TX_INT(RST_USART_TX_INT)
	`endif
		, .DO_ubrrl(DO_ubrrl)
		, .DO_ubrrh(DO_ubrrh)
		, .DO_ucsrc(DO_ucsrc)
		, .DO_ucsrb(DO_ucsrb)
		, .DO_ucsra(DO_ucsra)
		, .DO_udr(DO_udr)

		, .XCK_DDR(XCK_DDR), .XCK_OUT(XCK_OUT), .XCK_IN(XCK_IN) 
		, .TX_OUT(TX_OUT), .RX_IN(RX_IN)
		, .TXEN(USART_TXEN), .XCKEN(USART_XCKEN)

		);
`endif
`ifdef IOR_HAVE_TWI
    rtavr_ior_twi #(.BASE_ADDR(`TWI_ADDR)) i_twi(
		.CLK(CLK), .RESET(RESET), .DI(DI)
		, .DO(twi_DO)
		, .WR(WR), .RD(RD), .ADDR(ADDR) 
		, .TWI_EN(TWI_EN)
		, .SDA_OUT(SDA_OUT), .SDA_IN(SDA_IN)
		, .SCL_OUT(SCL_OUT), .SCL_IN(SCL_IN)
		, .TWI_SLAVE(TWI_SLAVE)
		, .RST_TWI_SLAVE(RST_TWI_SLAVE)
		);
`endif
`ifdef IOR_HAVE_CPI1
    rtavr_ior_cpi #(.BASE_ADDR(`CPI1_ADDR)) i_cpi1 (
		.RESET(RESET)
	`ifdef SYS_DRIVEN_CLK4PH
		, .CLK_0(CLK_0)
		, .CLK_180(CLK_180)
	`else // SYS_DRIVEN_CLK2X
		, .CLK2X(CLK2X)
		, .CLK(CLK)
	`endif
	`ifdef IOR_DUALPORT
		, .DI(DI_cpi1)
		, .WE_cpcsr(DP_WE_cpcsr1)
		, .WE_cpdr(DP_WE_cpdr1)
		, .OE_cpdr(DP_OE_cpdr1)
		// interrupt
		, .CPI_INT(DP_CPI1_INT)
		, .RST_CPI_INT(DP_RST_CPI1_INT)
	`else
		, .DI(DI)
		, .WE_cpcsr(WE_cpcsr1)
		, .WE_cpdr(WE_cpdr1)
		, .OE_cpdr(OE_cpdr1)
		// interrupt
		, .CPI_INT(CPI1_INT)
		, .RST_CPI_INT(RST_CPI1_INT)
	`endif
		, .DO_cpcsr(DO_cpcsr1)
		, .DO_cpdr(DO_cpdr1)
		// external I/O
		, .CPI_TXD(CPI1_TXD[7:0])
		, .CPI_TX(CPI1_TX)
		, .CPI_TX_ACK(CPI1_TX_ACK)
		, .CPI_IXS(CPI1_IXS)
		, .CPI_IXS_ACK(CPI1_IXS_ACK)

		, .CPI_RXD(CPI1_RXD[7:0])
		, .CPI_RX(CPI1_RX)
		, .CPI_RX_ACK(CPI1_RX_ACK)
		, .CPI_IXR(CPI1_IXR)
		, .CPI_IXR_ACK(CPI1_IXR_ACK)
		);
`endif
`ifdef IOR_HAVE_PORTA
    rtavr_ior_port #(.BASE_ADDR(`PORTA_ADDR)) i_port_a(
		.RESET(RESET)
	`ifdef SYS_DRIVEN_CLK4PH
		, .CLK_0(CLK_0)
		, .CLK_180(CLK_180)
	`else // SYS_DRIVEN_CLK2X
		, .CLK2X(CLK2X)
		, .CLK(CLK)
	`endif
	`ifdef IOR_DUALPORT
		, .DI_ddr(DI_ddr_a)
		, .DI_pin(DI_pin_a)
		, .DI_port(DI_port_a)
		, .WE_ddr(DP_WE_ddr_a)
		, .WE_pin(DP_WE_pin_a)
		, .WE_port(DP_WE_port_a)
	`else
		, .DI_ddr(DI)
		, .DI_pin(DI)
		, .DI_port(DI)
		, .WE_ddr(WE_ddr_a)
		, .WE_pin(WE_pin_a)
		, .WE_port(WE_port_a)
	`endif
		, .DO_ddr(DO_ddr_a)
		, .DO_pin(DO_pin_a)
		, .DO_port(DO_port_a)
		, .PIN(PIN[7:0]), .PORT(PORT_IN[7:0]), .DDR(DDR[7:0])
		);
`endif
`ifdef IOR_HAVE_PORTB
    rtavr_ior_port #(.BASE_ADDR(`PORTB_ADDR)) i_port_b(
		.RESET(RESET)
	`ifdef SYS_DRIVEN_CLK4PH
		, .CLK_0(CLK_0)
		, .CLK_180(CLK_180)
	`else // SYS_DRIVEN_CLK2X
		, .CLK2X(CLK2X)
		, .CLK(CLK)
	`endif
	`ifdef IOR_DUALPORT
		, .DI_ddr(DI_ddr_b)
		, .DI_pin(DI_pin_b)
		, .DI_port(DI_port_b)
		, .WE_ddr(DP_WE_ddr_b)
		, .WE_pin(DP_WE_pin_b)
		, .WE_port(DP_WE_port_b)
	`else
		, .DI_ddr(DI)
		, .DI_pin(DI)
		, .DI_port(DI)
		, .WE_ddr(WE_ddr_b)
		, .WE_pin(WE_pin_b)
		, .WE_port(WE_port_b)
	`endif
		, .DO_ddr(DO_ddr_b)
		, .DO_pin(DO_pin_b)
		, .DO_port(DO_port_b)
		, .PIN(PIN[15:8]), .PORT(PORT_IN[15:8]), .DDR(DDR[15:8])
		);
`endif
`ifdef IOR_HAVE_PORTC
    rtavr_ior_port #(.BASE_ADDR(`PORTC_ADDR)) i_port_c(
		.RESET(RESET)
	`ifdef SYS_DRIVEN_CLK4PH
		, .CLK_0(CLK_0)
		, .CLK_180(CLK_180)
	`else // SYS_DRIVEN_CLK2X
		, .CLK2X(CLK2X)
		, .CLK(CLK)
	`endif
	`ifdef IOR_DUALPORT
		, .DI_ddr(DI_ddr_c)
		, .DI_pin(DI_pin_c)
		, .DI_port(DI_port_c)
		, .WE_ddr(DP_WE_ddr_c)
		, .WE_pin(DP_WE_pin_c)
		, .WE_port(DP_WE_port_c)
	`else
		, .DI_ddr(DI)
		, .DI_pin(DI)
		, .DI_port(DI)
		, .WE_ddr(WE_ddr_c)
		, .WE_pin(WE_pin_c)
		, .WE_port(WE_port_c)
	`endif
		, .DO_ddr(DO_ddr_c)
		, .DO_pin(DO_pin_c)
		, .DO_port(DO_port_c)
		, .PIN(PIN[23:16]), .PORT(PORT_IN[23:16]), .DDR(DDR[23:16])
		);
`endif

`ifdef IOR_DUALPORT
   `ifdef SUPPORT_LOCKING
		reg lk2x;
		reg lk_last;

		wire lk_conflict = CPU0_LK_REQ & CPU1_LK_REQ & 
				( CPU0_LK_ADDR == CPU1_LK_ADDR ) & (
      `ifdef ZZZ
				(CPU0_LK_ADDR == (`RAMWIN_ADDR +1))	// RAMDR
	`ifdef IOR_HAVE_PORTA
				| (CPU0_LK_ADDR == `PORTA_ADDR)
				| (CPU0_LK_ADDR == `PORTA_ADDR +1)
				| (CPU0_LK_ADDR ==  `PORTA_ADDR +2)
	`endif
	`ifdef IOR_HAVE_PORTB
				| (CPU0_LK_ADDR == `PORTB_ADDR)
				| (CPU0_LK_ADDR == `PORTB_ADDR +1)
				| (CPU0_LK_ADDR ==  `PORTB_ADDR +2)
	`endif
	`ifdef IOR_HAVE_PORTC
				| (CPU0_LK_ADDR == `PORTC_ADDR)
				| (CPU0_LK_ADDR == `PORTC_ADDR +1)
				| (CPU0_LK_ADDR ==  `PORTC_ADDR +2)
	`endif
	`ifdef IOR_HAVE_GPIOR0
				| (CPU0_LK_ADDR == `GPIOR0_ADDR)
	`endif
	`ifdef IOR_HAVE_GPIOR1
				| (CPU0_LK_ADDR == `GPIOR1_ADDR)
	`endif
	`ifdef IOR_HAVE_GPIOR2
				| (CPU0_LK_ADDR == `GPIOR2_ADDR)
	`endif
      `else // ZZZ
				~ CPU0_LK_ADDR[5]
      `endif

				);
		wire winner = ~lk2x ^ lk_last;
		assign CPU0_LK_NAK = lk_conflict & winner;
		assign CPU1_LK_NAK = lk_conflict & ~winner;

		always @(`posedge_CLK_270)
		begin
		 if (RESET)
		    begin
			lk2x       <= 0;
			lk_last    <= 0;
		    end
		 else if (~__CLK_270)
		    begin
			lk2x       <=  (CPU0_LK_REQ2X | CPU1_LK_REQ2X);
			if (lk_conflict) lk_last    <=  winner;
		    end
		end
//		assign CPU0_LK_NAK = 1'b0;
//		assign CPU1_LK_NAK = 1'b0;
   `endif // SUPPORT_LOCKING
`endif

  generate 
    genvar i;
  `ifdef IOR_HAVE_PORTA
    for (i=0; i<8; i=i+1)
       begin : port_a_out
	assign PORT[i] = 
    `ifdef IOR_HAVE_TIM0_OC0A
		(`OC0A_BIT == i) & OC0A_EN ? OC0A_OUT :
        `ifdef IOR_DUALPORT
		(`OC0A_BIT == i) & PS1_OC0A_EN ? PS1_OC0A_OUT :
        `endif
    `endif
    `ifdef IOR_HAVE_TIM0_OC0B
		(`OC0B_BIT == i) & OC0B_EN ? OC0B_OUT :
        `ifdef IOR_DUALPORT
		(`OC0A_BIT == i) & PS1_OC0B_EN ? PS1_OC0B_OUT :
        `endif
    `endif
    `ifdef IOR_HAVE_SPI
		(`SCK_BIT == i) & SPI_MASTER ? SCK_OUT :
		(`MOSI_BIT == i) & SPI_MASTER ? MOSI_OUT :
		(`MISO_BIT == i) & SPI_SLAVE ? MISO_OUT :
    `endif
    `ifdef IOR_HAVE_USART
		(`XCK_BIT == i) & USART_XCKEN ? XCK_OUT :
		(`TX_BIT == i) & USART_TXEN ? TX_OUT :
    `endif
    `ifdef IOR_HAVE_TWI
		(`SDA_BIT == i) & TWI_EN & ~SDA_OUT ? 0 :
		(`SCL_BIT == i) & TWI_EN & ~SCL_OUT ? 0 :
    `endif
		PORT_IN[i];
       end
  `endif
  `ifdef IOR_HAVE_PORTB
    for (i=8; i<16; i=i+1)
       begin : port_b_out
	assign PORT[i] = 
    `ifdef IOR_HAVE_TIM0_OC0A
		(`OC0A_BIT == i) & OC0A_EN ? OC0A_OUT :
        `ifdef IOR_DUALPORT
		(`OC0A_BIT == i) & PS1_OC0A_EN ? PS1_OC0A_OUT :
        `endif
    `endif
    `ifdef IOR_HAVE_TIM0_OC0B
		(`OC0B_BIT == i) & OC0B_EN ? OC0B_OUT :
        `ifdef IOR_DUALPORT
		(`OC0A_BIT == i) & PS1_OC0B_EN ? PS1_OC0B_OUT :
        `endif
    `endif
    `ifdef IOR_HAVE_SPI
		(`SCK_BIT == i) & SPI_MASTER ? SCK_OUT :
		(`MOSI_BIT == i) & SPI_MASTER ? MOSI_OUT :
		(`MISO_BIT == i) & SPI_SLAVE ? MISO_OUT :
    `endif
    `ifdef IOR_HAVE_USART
		(`XCK_BIT == i) & USART_XCKEN ? XCK_OUT :
		(`TX_BIT == i) & USART_TXEN ? TX_OUT :
    `endif
    `ifdef IOR_HAVE_TWI
		(`SDA_BIT == i) & TWI_EN & ~SDA_OUT ? 0 :
		(`SCL_BIT == i) & TWI_EN & ~SCL_OUT ? 0 :
    `endif
		PORT_IN[i];
       end
  `endif
  `ifdef IOR_HAVE_PORTC
    for (i=16; i<24; i=i+1)
       begin : port_c_out
	assign PORT[i] = 
    `ifdef IOR_HAVE_TIM0_OC0A
		(`OC0A_BIT == i) & OC0A_EN ? OC0A_OUT :
        `ifdef IOR_DUALPORT
		(`OC0A_BIT == i) & PS1_OC0A_EN ? PS1_OC0A_OUT :
        `endif
    `endif
    `ifdef IOR_HAVE_TIM0_OC0B
		(`OC0B_BIT == i) & OC0B_EN ? OC0B_OUT :
        `ifdef IOR_DUALPORT
		(`OC0A_BIT == i) & PS1_OC0B_EN ? PS1_OC0B_OUT :
        `endif
    `endif
    `ifdef IOR_HAVE_SPI
		(`SCK_BIT == i) & SPI_MASTER ? SCK_OUT :
		(`MOSI_BIT == i) & SPI_MASTER ? MOSI_OUT :
		(`MISO_BIT == i) & SPI_SLAVE ? MISO_OUT :
    `endif
    `ifdef IOR_HAVE_USART
		(`XCK_BIT == i) & USART_XCKEN ? XCK_OUT :
		(`TX_BIT == i) & USART_TXEN ? TX_OUT :
    `endif
    `ifdef IOR_HAVE_TWI
		(`SDA_BIT == i) & TWI_EN & ~SDA_OUT ? 0 :
		(`SCL_BIT == i) & TWI_EN & ~SCL_OUT ? 0 :
    `endif
		PORT_IN[i];
       end
  `endif

  endgenerate

  `ifdef IOR_HAVE_SPI
	assign SS_DDR = DDR[`SS_BIT];
	assign SS_IN = PIN[`SS_BIT];
	assign SCK_IN = PIN[`SCK_BIT];
	assign MOSI_IN = PIN[`MOSI_BIT];
	assign MISO_IN = PIN[`MISO_BIT];
  `endif
  `ifdef IOR_HAVE_USART
	assign XCK_DDR = DDR[`XCK_BIT];
	assign XCK_IN = PIN[`XCK_BIT];
	assign RX_IN = PIN[`RX_BIT];
  `endif
  `ifdef IOR_HAVE_TWI
	assign SDA_IN = PIN[`SDA_BIT];
	assign SCL_IN = PIN[`SCL_BIT];
  `endif
  `ifdef IOR_HAVE_INT0
	assign PS0_PIN_INT0 = PIN[`INT0_BIT];
    `ifdef IOR_HAVE_INT1
	assign PS0_PIN_INT1 = PIN[`INT1_BIT];
    `endif
    `ifdef IOR_DUALPORT
	assign PS1_PIN_INT0 = PIN[`INT0_BIT];
      `ifdef IOR_HAVE_INT1
	assign PS1_PIN_INT1 = PIN[`INT1_BIT];
      `endif
    `endif
  `endif
endmodule
