/*
 * Copyright (C) 2011 Koji Suzuki
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.1 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

`include "rtavr_defs.v"
// global defines : 

//`define IOR_HAVE_CPI0

//`define IOR_HAVE_INT0
//`define IOR_HAVE_TIM0
//`define IOR_HAVE_TIM0_OC0A
//`define IOR_HAVE_TIM0_OC0B
//`define IOR_HAVE_SREG_SP

// Module Name: rtavr_ior_ps
// 
// Target Device:   xc3s200a-4ft256
// Product Version: ISE 12.4
// Design Goal:     Balanced
//

module rtavr_ior_ps # (
		parameter CPUID = 0
) (
		input RESET,
		input WR,RD, 
		input [7:0] DI,
		input [5:0] ADDR,
		output [7:0] DO,
		output ACK
`ifdef SYS_DRIVEN_CLK4PH
		, input CLK_0
		, input CLK_180
		, input CLK_270
`else // SYS_DRIVEN_CLK2X
		, input CLK2X
		, input CLK
`endif
`ifdef IOR_HAVE_SREG_SP
		, input  [7:0] FLAGS_IN
		, input  [15:0] SP_IN
		, output [7:0] FLAGS_OUT
		, output [15:0] SP_OUT
`endif
		// interrupt handling
`ifdef IOR_HAVE_INT0
		, input  PIN_INT0
		, input  RST_INT0
		, output INT0
  `ifdef IOR_HAVE_INT1
		, input  PIN_INT1
		, input  RST_INT1
		, output INT1
  `endif
`endif
`ifdef IOR_HAVE_RAMWIN
		, output RAMW_WE
		, output RAMW_OE
		, output [12:0] RAMW_ADDR
		, input [7:0] RAMW_DO
`endif
`ifdef IOR_HAVE_NVM
		, output ROM_WP
		, output RAM_WP
`endif

`ifdef IOR_HAVE_TIM0
		, input  RST_TIM0_OVF
		, output TIM0_OVF
		, input  RST_TIM0_COMPA	
		, output TIM0_COMPA	
		, input  RST_TIM0_COMPB
		, output TIM0_COMPB
    `ifdef IOR_HAVE_TIM0_OC0A
                , output OC0A_EN
                , output OC0A_OUT
    `endif
    `ifdef IOR_HAVE_TIM0_OC0B
                , output OC0B_EN
                , output OC0B_OUT
    `endif
`endif
`ifdef IOR_HAVE_CPI0
		// external I/O
		, output [7:0] CPI0_TXD
		, output CPI0_TX
		, input CPI0_TX_ACK
		, output CPI0_IXS
		, input CPI0_IXS_ACK

		, input [7:0] CPI0_RXD
		, input CPI0_RX
		, output CPI0_RX_ACK
		, input CPI0_IXR
		, output CPI0_IXR_ACK

		// interrupt
		, input RST_CPI0_INT
		, output CPI0_INT
`endif
		);
  
`ifdef SYS_DRIVEN_CLK4PH
//    wire CLK = CLK_180;
//    wire __CLK_90 =  1'b0;
    wire __CLK_180 = 1'b0;
    wire __CLK_270 =  1'b0;
`else // SYS_DRIVEN_CLK2X
//    wire __CLK_90 = CLK;
    wire __CLK_180 = CLK;
    wire __CLK_270 = ~CLK;
`endif

//`define	 RAMWIN_ADDR 6'h1f   // RAMDR 0x1f , RAMAR 0x20
//`define	 TIM0_ADDR   6'h15   // OCR0B 0x15 - TCCR0A 0x19
//`define	 SREG_ADDR   6'h3f   // SREG 0x3f , SPH,SPL,CCP,RSRFR,MCUCR 0x3a
//`define	 MCUCR_ADDR  6'h3a   // MCUCR 0x3a
//`define	 GIFR_ADDR   6'h0b   // GIFR 0x0b , GIMSK 0x0c
//`define	 TIFR_ADDR   6'h25   // TIFR 0x25 , TIMSK 0x26  
//`define	 NVM_ADDR    6'h32   // NVMCSR 0x32 , NVMCMD 0x33  

    wire  [7:0] DO_async;
    assign DO[7:0] = DO_async[7:0];

`ifdef IOR_HAVE_SREG_SP
    reg  [7:0] r_sreg;
    reg  [7:0] r_sph;
    reg  [7:0] r_spl;

    assign FLAGS_OUT = r_sreg;
    assign SP_OUT  = { r_sph , r_spl };
    wire WE_sreg = WR & (ADDR == `SREG_ADDR);
    wire WE_sph = WR & (ADDR == (`SREG_ADDR -1));
    wire WE_spl = WR & (ADDR == (`SREG_ADDR -2));
`endif

`ifdef IOR_HAVE_RAMWIN
  `ifdef IOR_HAVE_RAMWIN_HA
    reg  [4:0] r_ramar_h = 0;
  `else
    wire  [4:0] r_ramar_h = 0;
  `endif
    reg  [7:0] r_ramar = 0;

  `ifdef IOR_HAVE_RAMWIN_HA
    wire WE_ramar_h = WR & (ADDR == (`RAMWIN_ADDR -1));
  `endif
    wire WE_ramdr   = WR & (ADDR == `RAMWIN_ADDR);
    wire OE_ramdr   = RD & (ADDR == `RAMWIN_ADDR); 
    wire WE_ramar   = WR & (ADDR == (`RAMWIN_ADDR +1));
     
    assign RAMW_ADDR[7:0] = r_ramar;
    assign RAMW_ADDR[12:8] = r_ramar_h;
    assign RAMW_WE = WE_ramdr;
    assign RAMW_OE = OE_ramdr;
    
    wire [7:0] DO_ramar_h = {3'b0 , r_ramar_h};
    wire [7:0] DO_ramar   = r_ramar;
`endif

`ifdef IOR_HAVE_NVM
    wire WE_nvmcmd = WR & (ADDR == (`NVM_ADDR +1));

    reg nvm_wp = 0;		// 0: protected 1: un-protected
    assign ROM_WP = nvm_wp;
   `ifdef NVM_RAM_PROTECT
	reg nvm_wp_ram = 0;
   `else
	wire nvm_wp_ram = 1;	// 0: protected 1: un-protected
   `endif
   assign RAM_WP = nvm_wp_ram;
   wire NVM_CPUID = CPUID;
   wire [7:0] DO_nvmcsr = { ~nvm_wp , ~nvm_wp_ram , 5'h0, NVM_CPUID };
`endif

`ifdef IOR_HAVE_TIM0
    wire WE_tifr = WR & (ADDR == `TIFR_ADDR);
    wire WE_timsk = WR & (ADDR == (`TIFR_ADDR +1));

    reg t0_ovf_prev;  
    reg t0_compa_prev;
    reg t0_compb_prev;  
    reg t0_ovf, t0_ovf_ie; // bit0
    reg t0_compa, t0_compa_ie; // bit1
    reg t0_compb, t0_compb_ie; // bit2

    assign TIM0_OVF   = t0_ovf & t0_ovf_ie;       // int 10
    assign TIM0_COMPA = t0_compa & t0_compa_ie ;  // int 11
    assign TIM0_COMPB = t0_compb & t0_compb_ie ;  // int 12
  `ifdef TIM0_EXPORT_PS
     wire [11:0] TIM0_PS;
  `endif
`endif

 //reg intf0, int0; // bit0
`ifdef IOR_HAVE_INT0
    wire WE_gifr = WR & (ADDR == `GIFR_ADDR);
    wire WE_gimsk = WR & (ADDR == (`GIFR_ADDR +1));
    wire WE_mcucr = WR & (ADDR == `MCUCR_ADDR);

    reg gi_int0_prev; 
    reg gi_int0, gi_int0_ie; // bit0
    reg [1:0] mcucr_isc;
  
    assign INT0 = gi_int0 & gi_int0_ie;
  `ifdef IOR_HAVE_INT1
      reg gi_int1_prev; 
      reg gi_int1, gi_int1_ie; // bit1
      reg [1:0] mcucr_isc1;
  
      assign INT1 = gi_int1 & gi_int1_ie;
  `else
      wire gi_int1 = 0;
      wire gi_int1_ie = 0;
      wire [1:0] mcucr_isc1 = 2'b00;
  `endif
`else // IOR_HAVE_INT0
    wire OE_gifr = 0; 
    wire WE_gifr = 0;
    wire OE_gimsk = 0; 
    wire WE_gimsk = 0;
    wire OE_mcucr = 0; 
    wire WE_mcucr = 0;
`endif

  `ifdef IOR_HAVE_CPI0
	wire WE_cpcsr0 = WR & (ADDR == `CPI0_ADDR);
	wire WE_cpdr0  = WR & (ADDR == (`CPI0_ADDR + 1));
//	wire OE_cpcsr0 = RD & (ADDR == `CPI0_ADDR);
	wire OE_cpdr0  = RD & (ADDR == (`CPI0_ADDR + 1)); 
	wire [7:0] DO_cpcsr0;
	wire [7:0] DO_cpdr0;
  `endif
  `ifdef IOR_HAVE_TIM0
	wire WE_ocr0b = WR & (ADDR == `TIM0_ADDR);
	wire WE_ocr0a = WR & (ADDR == (`TIM0_ADDR + 1));
	wire WE_tcnt0 = WR & (ADDR == (`TIM0_ADDR + 2));
	wire WE_tccr0b = WR & (ADDR == (`TIM0_ADDR + 3));
	wire WE_tccr0a = WR & (ADDR == (`TIM0_ADDR + 4));
	wire [7:0] DO_ocr0b;
	wire [7:0] DO_ocr0a;
	wire [7:0] DO_tcnt0;
	wire [7:0] DO_tccr0b;
	wire [7:0] DO_tccr0a;
	wire [7:0] DO_tifr = {5'b0 , t0_compb, t0_compa ,t0_ovf};
	wire [7:0] DO_timsk = {5'b0 , t0_compb_ie, t0_compa_ie ,t0_ovf_ie};
  `endif
  `ifdef IOR_HAVE_INT0
	wire [7:0] DO_gifr  = {6'b0 , gi_int1 , gi_int0};
	wire [7:0] DO_gimsk = {6'b0, gi_int1_ie, gi_int0_ie};
	wire [7:0] DO_mcucr = {mcucr_isc , mcucr_isc1 , 4'b0 };
  `endif
  
  assign DO_async = ~RD ? 8'h00
	: (ADDR == `SREG_ADDR) ?      r_sreg 
	: (ADDR == (`SREG_ADDR -1)) ? r_sph
	: (ADDR == (`SREG_ADDR -2)) ? r_spl
  `ifdef IOR_HAVE_CPI0
	: (ADDR == `CPI0_ADDR) ?       DO_cpcsr0
	: (ADDR == (`CPI0_ADDR + 1)) ? DO_cpdr0
  `endif
  `ifdef IOR_HAVE_TIM0
	: (ADDR == `TIM0_ADDR) ?       DO_ocr0b
	: (ADDR == (`TIM0_ADDR + 1)) ? DO_ocr0a
	: (ADDR == (`TIM0_ADDR + 2)) ? DO_tcnt0
	: (ADDR == (`TIM0_ADDR + 3)) ? DO_tccr0b
	: (ADDR == (`TIM0_ADDR + 4)) ? DO_tccr0a
	: (ADDR == `TIFR_ADDR) ?       DO_tifr
	: (ADDR == (`TIFR_ADDR +1)) ?  DO_timsk
  `endif
  `ifdef IOR_HAVE_RAMWIN
 	: (ADDR == (`RAMWIN_ADDR -1)) ? DO_ramar_h
//	: (ADDR == `RAMWIN_ADDR)      ? RAMW_DO
	: (ADDR == (`RAMWIN_ADDR +1)) ? DO_ramar
  `endif
  `ifdef IOR_HAVE_NVM
	: (ADDR == `NVM_ADDR) ? 	DO_nvmcsr
  `endif
  `ifdef IOR_HAVE_INT0
	: (ADDR == `GIFR_ADDR) ?        DO_gifr
	: (ADDR == (`GIFR_ADDR +1)) ?   DO_gimsk
	: (ADDR == `MCUCR_ADDR) ?       DO_mcucr
  `endif
	: 8'h00;

  assign ACK =
	  (ADDR == `SREG_ADDR)
	| (ADDR == (`SREG_ADDR -1))
	| (ADDR == (`SREG_ADDR -2))
  `ifdef IOR_HAVE_CPI0
	| (ADDR == `CPI0_ADDR)
	| (ADDR == (`CPI0_ADDR + 1))
  `endif
  `ifdef IOR_HAVE_TIM0
	| (ADDR == `TIM0_ADDR)
	| (ADDR == (`TIM0_ADDR + 1))
	| (ADDR == (`TIM0_ADDR + 2))
	| (ADDR == (`TIM0_ADDR + 3))
	| (ADDR == (`TIM0_ADDR + 4))
	| (ADDR == `TIFR_ADDR)
	| (ADDR == (`TIFR_ADDR +1))
  `endif
  `ifdef IOR_HAVE_RAMWIN
 	| (ADDR == (`RAMWIN_ADDR -1))
	| (ADDR == `RAMWIN_ADDR)
	| (ADDR == (`RAMWIN_ADDR +1))
  `endif
  `ifdef IOR_HAVE_NVM
	| (ADDR == `NVM_ADDR)
	| (ADDR == (`NVM_ADDR + 1))
  `endif
  `ifdef IOR_HAVE_INT0
	| (ADDR == `GIFR_ADDR)
	| (ADDR == (`GIFR_ADDR +1))
	| (ADDR == `MCUCR_ADDR)
  `endif
	;

`ifdef IOR_HAVE_CPI0
    rtavr_ior_cpi #(.BASE_ADDR(`CPI0_ADDR)) i_cpi0 (
		.RESET(RESET), .DI(DI)
  `ifdef SYS_DRIVEN_CLK4PH
		, .CLK_0(CLK_0)
		, .CLK_180(CLK_180)
  `else // SYS_DRIVEN_CLK2X
		, .CLK2X(CLK2X)
		, .CLK(CLK)
  `endif
		, .WE_cpcsr(WE_cpcsr0)
		, .WE_cpdr(WE_cpdr0)
		, .OE_cpdr(OE_cpdr0)
		, .DO_cpcsr(DO_cpcsr0)
		, .DO_cpdr(DO_cpdr0)
		// external I/O
		, .CPI_TXD(CPI0_TXD[7:0])
		, .CPI_TX(CPI0_TX)
		, .CPI_TX_ACK(CPI0_TX_ACK)
		, .CPI_IXS(CPI0_IXS)
		, .CPI_IXS_ACK(CPI0_IXS_ACK)

		, .CPI_RXD(CPI0_RXD[7:0])
		, .CPI_RX(CPI0_RX)
		, .CPI_RX_ACK(CPI0_RX_ACK)
		, .CPI_IXR(CPI0_IXR)
		, .CPI_IXR_ACK(CPI0_IXR_ACK)

		// interrupt
		, .CPI_INT(CPI0_INT)
		, .RST_CPI_INT(RST_CPI0_INT)
		);
`endif
`ifdef IOR_HAVE_TIM0
    rtavr_ior_timer0 #(.BASE_ADDR(`TIM0_ADDR)) i_timer0(
		.RESET(RESET), .DI(DI)
`ifdef SYS_DRIVEN_CLK4PH
		, .CLK_0(CLK_0)
		, .CLK_180(CLK_180)
`else // SYS_DRIVEN_CLK2X
		, .CLK2X(CLK2X)
		, .CLK(CLK)
`endif
		, .WE_ocr0b(WE_ocr0b)
		, .WE_ocr0a(WE_ocr0a)
		, .WE_tcnt0(WE_tcnt0)
		, .WE_tccr0b(WE_tccr0b)
		, .WE_tccr0a(WE_tccr0a)
		, .DO_ocr0b(DO_ocr0b)
		, .DO_ocr0a(DO_ocr0a)
		, .DO_tcnt0(DO_tcnt0)
		, .DO_tccr0b(DO_tccr0b)
		, .DO_tccr0a(DO_tccr0a)
`ifdef IOR_HAVE_TIM0_OC0A
		, .OC0A_EN(OC0A_EN)
		, .OC0A_OUT(OC0A_OUT)
`endif
`ifdef IOR_HAVE_TIM0_OC0B
		, .OC0B_EN(OC0B_EN)
		, .OC0B_OUT(OC0B_OUT)
`endif
		, .TIM0_OVF(TIM0_OVF_IN)
		, .TIM0_COMPB(TIM0_COMPB_IN)
		, .TIM0_COMPA(TIM0_COMPA_IN)
    `ifdef TIM0_EXPORT_PS
		, .PS(TIM0_PS)
    `endif
		 );
`endif

`ifdef IOR_HAVE_SREG_SP
    always @(`posedge_CLK_270)
    begin
      if (RESET)
        begin
	r_sreg <= 0;
  `ifdef FIXED_SPH
	r_sph <= `FIXED_SPH;
  `else
	r_sph <= 0;
  `endif
	r_spl <= 0;
        end
      else if (~__CLK_270)
        begin
	if(WE_sreg) r_sreg <= DI;
	else r_sreg <= FLAGS_IN;

  `ifdef FIXED_SPH
  `else
    `ifdef IOR_SPHBITS
	if(WE_sph) r_sph <= DI[`IOR_SPHBITS-1:0];
	else r_sph <= SP_IN[`IOR_SPHBITS+7:8];
    `else
	if(WE_sph) r_sph <= DI;
	else r_sph <= SP_IN[15:8];
    `endif
  `endif
	if(WE_spl) r_spl <= DI;
	else r_spl <= SP_IN[7:0];
        end
    end
`endif

  always @(`posedge_CLK_180)
  begin
    if (RESET)
      begin
`ifdef IOR_HAVE_TIM0
	t0_compb_ie <= 0;
	t0_compa_ie <= 0;
	t0_ovf_ie <= 0;

	t0_ovf <= 0;
	t0_compa <= 0;

	t0_compb <= 0;

	t0_ovf_prev   <=  0;
	t0_compa_prev <=  0;
	t0_compb_prev <=  0;
`endif

`ifdef IOR_HAVE_INT0
	mcucr_isc <= 0;

	gi_int0_ie <= 0;
	gi_int0 <= 0;
	gi_int0_prev   <= 0;
  `ifdef IOR_HAVE_INT1
	mcucr_isc1 <= 0;

	gi_int1_ie <= 0;
	gi_int1 <= 0;
	gi_int1_prev   <= 0;
  `endif
`endif
      end
    else if (~__CLK_180) // posedge CLK_180
      begin
`ifdef IOR_HAVE_TIM0
	if(WE_timsk) {t0_compb_ie, t0_compa_ie ,t0_ovf_ie} <= DI[2:0];

	if(RST_TIM0_OVF) t0_ovf <= 0;
	else if((WE_tifr & DI[0]) 
             | (t0_ovf_prev ^ TIM0_OVF_IN) & ~t0_ovf_prev ) t0_ovf <= 1;

	if(RST_TIM0_COMPA) t0_compa <= 0;
	else if((WE_tifr & DI[1])
             | (t0_compa_prev ^ TIM0_COMPA_IN) & ~t0_compa_prev) t0_compa <= 1;

	if(RST_TIM0_COMPB) t0_compb <= 0;
	else if((WE_tifr & DI[2])
             | (t0_compb_prev ^ TIM0_COMPB_IN) & ~t0_compb_prev) t0_compb <= 1;

	t0_ovf_prev   <=  TIM0_OVF_IN;
	t0_compa_prev <=  TIM0_COMPA_IN;
	t0_compb_prev <=  TIM0_COMPB_IN; 
`endif //  IOR_HAVE_TIM0

`ifdef IOR_HAVE_INT0
	if(WE_mcucr) {mcucr_isc} <= DI[7:6];

	if(WE_gimsk) {gi_int0_ie} <= DI[0];
	if(RST_INT0 | (WE_gifr & DI[0])) gi_int0 <= 0;
	else if((gi_int0_prev ^ PIN_INT0)
                 & ((mcucr_isc == 1) 
                   | (mcucr_isc == 2) & ( gi_int0_prev) 
                   | (mcucr_isc == 3) & ~( gi_int0_prev) )  
	       | ((mcucr_isc == 0) & ~PIN_INT0) ) gi_int0 <= 1;

	gi_int0_prev   <= PIN_INT0;
  `ifdef IOR_HAVE_INT1
	if(WE_mcucr) {mcucr_isc1} <= DI[5:4];

	if(WE_gimsk) {gi_int1_ie} <= DI[1];
	if(RST_INT1 | (WE_gifr & DI[1])) gi_int1 <= 0;
	else if((gi_int1_prev ^ PIN_INT1)
                 & ((mcucr_isc1 == 1) 
                   | (mcucr_isc1 == 2) & ( gi_int1_prev) 
                   | (mcucr_isc1 == 3) & ~( gi_int1_prev) )  
	       | ((mcucr_isc1 == 0) & ~PIN_INT1) ) gi_int1 <= 1;

	gi_int1_prev   <= PIN_INT1;
  `endif
`endif // IOR_HAVE_INT0
      end
  end

  always @(`posedge_CLK_180)
  if (~__CLK_180) // posedge CLK_180
      begin
`ifdef IOR_HAVE_RAMWIN
	if(WE_ramar) r_ramar	   <= DI[7:0];
  `ifdef IOR_HAVE_RAMWIN_HA
	if(WE_ramar_h) r_ramar_h   <= DI[4:0];
  `endif
`endif
`ifdef IOR_HAVE_NVM
	if (WE_nvmcmd) nvm_wp     <= ( DI[5:0] == 6'h1d );
    `ifdef NVM_RAM_PROTECT
	if (WE_nvmcmd) nvm_wp_ram <= ( DI[5:0] == 6'h2d );
    `endif
`endif
      end

endmodule
