#include <stdlib.h>
#include <string.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>
#include <util/delay.h>

#define bit_set(port,bit)       { port |= (1<<(bit));}
#define bit_clear(port,bit)     { port &= ~(1<<(bit));}
#define min(a,b)                ((a)<(b) ? (a):(b))

#define nop()   __asm__ volatile ("nop" )
#define wdr()   __asm__ volatile("wdr" )
#define sleep()   __asm__ volatile("sleep" )

#ifndef GPIOR0
#define GPIOR0 _SFR_IO8(0x03)
#endif

#ifndef RAMARH
#define RAMARH _SFR_IO8(0x1e)
#endif

#ifndef RAMAR
#define RAMAR _SFR_IO8(0x20)
#endif

#ifndef RAMDR
#define RAMDR _SFR_IO8(0x1f)
#endif

char lock;

int8_t t_lock(uint8_t val) {
    uint8_t __result;

   RAMARH = (uint16_t)(&lock) >> 8;
   RAMAR =  (uint16_t)(&lock) & 0xff;

    __asm__ (
        "sbis 0x1f, 7"  "\n\t"
        "out  0x1f, %1" "\n\t"
        "in   %0, 0x1f" "\n\t"
        : "=r" (__result)
        : "r" (val)
    );
    return __result;
}

void t_unlock() {
	lock = 0;
}

int8_t t_test_and_set() {
    uint8_t __result;
    __asm__ (
        "clr %0" "\n\t"
        "sbi 0x03, 7" "\n\t"
        "bld %0, 0"   "\n\t"
        "bst %0, 1"   "\n\t"
        : "=r" (__result)
    );
    return __result;
}

int8_t t_test_and_clear() {
    uint8_t __result;
    __asm__ (
        "clr %0"      "\n\t"
        "cbi 0x03, 6" "\n\t"
        "bld %0, 0"   "\n\t"
        "bst %0, 1"   "\n\t"
        : "=r" (__result)
    );
    return __result;
}
